// Persistence of Vision Ray Tracer Scene Description File
// File: example_2.pov
// Vers: 3.6.1
// Desc: Topview of Tree for Surcoat-Macro 
// Date: 09.12.04 (dd.mm.yy)
// Auth: Tim Nikias Wenclawiak - www.NoLights.de
// Last Update: 07.01.05 (dd.mm.yy)

//Required Files
// "surcoat.inc" - The Surcoat-Macros
// "arial.ttf"   - TrueTypeFont to add some description text

//Description
// This scene will sample a test object and show two sorting-macros at work.
// 1.
// In the the upper left corner is the unmodified sampling. Since we're looking
// at the object along the same direction as the coating is sampled, we can
// see nothing of the test object.
// 2.
// In the upper right corner, the "roofed" samples were removed. The technique
// for this is to place a disc above the sample and shoot various samples at it.
// Depending on how many samples reach the disc without intersecting the given
// object, a sample is considered "roofed" and will be removed.
// 3.
// In the lower left corner, the "angled" samples are removed. Given a Normal
// and a maximum angle in degrees, the macro will sort through the samples
// and check the surface-normal of the sample in relation with the given normal.
// If the angle is higher than the given one, the sample is removed.
// 4.
// In the lower right corner, everything inside the transparent box was kept.

// Note that two other sorting macros aren't presented here. The first is too
// difficult to visualize and would require a mesh, which tend to be fairly
// large, the second is straight-forward and shouldn't need any further explanation:

// Surcoat_KeepMesh(Mesh_Object,ModFactor)
//  - This macro counts the amount of surfaces intersected when tracing from the
//    sample out of the mesh. If mod(amount,2)=ModFactor, the sample is kept.
//    In essence, if ModFactor is 0, all samples on theoretical outer surfaces
//    of the mesh are kept, if 1, all samples on inside are kept.
//    This macro is especially of use if you're using meshes created with other
//    applications, as they sometimes tend to overlap surfaces.

// Surcoat_RemoveOnObject(An_Object,Radius)
//  - If a ray, traced through the sample along the sample's normal, hits the given
//    object within the given radius, the sample is removed.

//
/*Commandline
+AM2 +A0.1 +R2 +W800 +H640
*/

#include "surcoat.inc"

global_settings{assumed_gamma 1}
#local Size = 25;
camera{
  orthographic
  location <0,0,-20>
  up y*Size
  right x*Size*1.33
  look_at <0,0,-10>
}

#declare Sun_Rotation = <125,-42,0>;

//Sunlight
light_source{<0,-1500,0>
  rgb 1
  rotate Sun_Rotation
  }

//Backlight
light_source{
  <-500,350,500>
  rgb <.05,.05,.3>
  shadowless
  parallel point_at <0,0,0>
}

//CSG-Object for tests
#declare Surcoat_Topview = pigment{image_map{png "example_1.png"}}

#declare Surcoat_Debug = on;
#declare Surcoat_Seed = 211083;

#declare Surcoat_CellResolution = 3;
#declare Surcoat_CellSamples = 15;

#declare Surcoat_Sample_Offset = .0001;
#declare Surcoat_Sample_Direction = -y-x+z;
#declare Surcoat_Sample_MaxDirectionDeviation = 90;
#declare Surcoat_Sample_OverrideDirection = 0;


#declare Test_Object = 
  merge{
    box{<-3,0,-3><3,1,3>}
    cylinder{<-2.5,0,-2.5><-2.5,4.01,-2.5>,.5}
    box{<-3,4,-3><3,5,3>}
    sphere{y*6,1}
    cylinder{y*4.5,y*6,1}
    pigment{rgb x}
  }    

#declare Boundary_Object =
  box{<-4,-1,-4>,<0,7.1,4>
    pigment{rgbt <.3,.3,1,.5>}
  }

//Create initial samples on surface
Surcoat_CreateSamples(Test_Object)

//Save to show unmodified samples
Surcoat_SaveSurface("base.tmp")

//The roofed samples
Surcoat_PrepareSorting()
Surcoat_RemoveRoofed(Test_Object,3.2,y,3,.65,45) //Object, Disc-Distance, -Normal, -Radius, Cover-Percentage to remove, Samples
Surcoat_SaveSorted("roof.tmp")

//Angled
Surcoat_PrepareSorting()
Surcoat_KeepAngled(y,60)
Surcoat_SaveSorted("angle.tmp")

//Keep inside of object
Surcoat_PrepareSorting()
Surcoat_KeepInside(Boundary_Object)
Surcoat_SaveSorted("bound.tmp")


//Display the samples
union{
  Surcoat_Spheres("base.tmp",.1)
  object{Test_Object}
  pigment{rgb 1}
  rotate y*45 rotate x*-25
  translate <-7,5,0>
}

union{
  Surcoat_Spheres("roof.tmp",.1)
  object{Test_Object}
  pigment{rgb 1}
  rotate y*45 rotate x*-25
  translate <7,5,0>
}

union{
  Surcoat_Spheres("angle.tmp",.1)
  object{Test_Object}
  pigment{rgb 1}
  rotate y*45 rotate x*-25
  translate <-7,-7,0>
}
union{
  Surcoat_Spheres("bound.tmp",.1)
  object{Test_Object}
  object{Boundary_Object}
  pigment{rgb 1}
  rotate y*45 rotate x*-25
  translate <7,-7,0>
}

//Description text
text{
  ttf "arial.ttf" "coated" .2,0
  pigment{rgb 1}
  finish{ambient 1 diffuse 0}
  scale 2
  translate <-10,1,0>
}

text{
  ttf "arial.ttf" "-roof" .2,0
  pigment{rgb 1}
  finish{ambient 1 diffuse 0}
  scale 2
  translate <5,1,0>
}

text{
  ttf "arial.ttf" "-angle" .2,0
  pigment{rgb 1}
  finish{ambient 1 diffuse 0}
  scale 2
  translate <-10,-11,0>
}

text{
  ttf "arial.ttf" "inside" .2,0
  pigment{rgb 1}
  finish{ambient 1 diffuse 0}
  scale 2
  translate <5,-11,0>
}

