// Persistence of Vision Ray Tracer Scene Description File
// File: sc2snow.inc
// Vers: 3.6.1
// Desc: Surcoat To Snow-Plug-In 
// Date: 17.12.04 (dd.mm.yy)
// Auth: Tim Nikias Wenclawiak - www.NoLights.de
// Last Update: 07.01.05 (dd.mm.yy)

//Required Files
// "surcoat.inc" - Surcoat Macros

//Description
// A set of macros designed to take input from my Surcoat-Macro-system and parse it to 
// create a layer of snow on objects.

//The main macro for this is Surcoat2Snowfile("Datafile","Savefile",Your_Object). It will
//load the given samples from "Datafile", process the samples and save to "Savefile". For
//various trace-calls, the object to test against has to be specified, which obviously
//should be the sampled-object itself.
//The macro works by testing what radius the sample may have and how high snow could stack
//at the sample's position. The results are then saved to disk and may be visualized with
//Snowfile_Display("Loadfile",Threshold,Radius,Strength,HeightModifier) which generates
//a blob using the data from "Loadfile". During tests I've found that the height of the snow
//often needs a little tweaking to get it right, and instead of reparsing the entire sample-
//set again, I just added a small modifier to quickly modify the height.

//There are various values which set limits for what the min and max size of a snow-particle is
//supposed to be, how high it can stack, how much influence walls have on the stacking height,
//etc.
//The only two parameters that really deserve a note are WallInfluence and StackMethod, as the
//others are pretty self-explanatory once you've looked at the example file and understood how
//the algorithm works on a basic level. WallInfluence is a multiplier for how far away a wall
//can be to affect the height of a snow-particle. If it is set to 1, walls don't affect particles
//anymore if they are further away than their maximum-height can be. Values larger than 1 stretch
//the effect more, whereas smaller values squish it so that only particles rather close to walls
//will stack snow.
//The StackMethod is actually a 0 or 1 switch to pick which test to use. Method 0 just memorizes
//the closest a wall gets, whereas Method 1 averages all distance-checks. Obviously, Method 1 has
//much smoother results but also less high stacking which often needs to be modified a little
//when displaying the snow.

//Then, there are three macros to adjust some aspects of the snow. Note that their name and
//parameters can not be changed, as they will be called from within another parsing-macro, namely
//Snowfile_Modify("Loadfile","Savefile") but they provide all data you can have access to.
//
//Without them, the snow would be spread and stacked *very* evenly, with little turbulence. To
//introduce this turbulence, two macros can be declared which affect the generated-dataset during
//the parse-time. They are Snowdata2Float_HeightAdd and Snowdata2Float_HeightMod which can be
//found just after the default values below. HeightAdd can be used to add some height to the stack
//and HeightMod modifies the generated height. Both can be generated using the supplied parameters,
//an example for this can be found in the example file.

//The third macro can be used to adjust the height-axis of the snow-particles. Normally,
//Surcoat_Snow_Sky would be used for this, but some objects might require slightly tilted stacking
//in certain areas (e.g. along the trunk of a tree). Again, using the parameters the macro
//is called with, you can derive a sky-vector appropriate for the particle.

#declare _sc2snow_inc_tnw = 1;
//Default-Values from Surcoat-Set
//The same as in "surcoat.inc"
#ifndef (Surcoat_Debug)
  #declare Surcoat_Debug = 0;
#end
#ifndef (Surcoat_Sample_Offset)
  #declare Surcoat_Sample_Offset = .0001;
#end

//Default values for this set
#declare Surcoat_Snow_Seed = 260981;
//"Sky-Vector" for the stacking
#declare Surcoat_Snow_Sky = y;
//Min and Max Radius for a snow-particle
#declare Surcoat_Snow_MinRadius = .1;
#declare Surcoat_Snow_MaxRadius = .3;
//Max-Stackingheight for a snow-particle
#declare Surcoat_Snow_MaxHeight = 1.5;
//Stackheight*WallInfluence is the distance until which snow-particles may stack
#declare Surcoat_Snow_WallInfluence = 2;
//Samples to determine the radius
#declare Surcoat_Snow_Samples_Radius = 10;
//Samples to determine proximity to walls
#declare Surcoat_Snow_Samples_Hemisphere = 50;
//After how many samples to debug a progression note
//(default is 0 and thus, "off")
#declare Surcoat_Snow_DebugSampleCount = 0;
//Default method is 0
#declare Surcoat_Snow_StackMethod = 0;

//Macro to generate Addition to Height
#macro Snowdata2Float_HeightAdd(_Position,_Normal,_Radius,_Height)
  //Calculate a float from the given data and declare it as
  // "Return". Default Macro returns 0.
  #local Return = 0;
  //At the end of the macro, return "Return", for simplicity's sake,
  //just keep this line. :o)
  (Return)
#end
//Macro to generate multiplier for the height
#macro Snowdata2Float_HeightMod(_Position,_Normal,_Radius,_Height)
  //Calculate a float from the given data and declare it as
  // "Return". Default Macro returns 1.
  #local Return = 1;
  //At the end of the macro, return "Return", for simplicity's sake,
  //just keep this line. :o)
  (Return)
#end
//Macro to dynamically adjust Sky-Vector for the stacking-cylinders
#macro Snowfile_Display_SkyMod(_Sky,_Pos,_Nor)
  //Calculate a new, normalized Sky-Vector from the given data...
  #local Return = _Sky;
  //...and return it.
  (Return)  
#end

#macro Surcoat2Snow_Defaults()
  //Both macros change nothing
  #macro Snowdata2Float_HeightAdd(_Position,_Normal,_Radius,_Height)
    #local Return = 0;
    (Return)
  #end
  #macro Snowdata2Float_HeightMod(_Position,_Normal,_Radius,_Height)
    #local Return = 1;
    (Return)
  #end
  #macro Snowfile_Display_SkyMod(_Sky,_Pos,_Nor)
    #local Return = _Sky;
    (Return)  
  #end
#end
Surcoat2Snow_Defaults()

//Parsing through each particle and saving properties
#macro Surcoat2Snowfile(_SnowInFile,_SnowOutFile,_Object)
  #if (Surcoat_Debug)
    #debug "\nCalculating snow-particle data...\n"
  #end
  #local _R=seed(Surcoat_Snow_Seed);
  //Shorter variable-names
  #local _Sky = vnormalize(Surcoat_Snow_Sky);
  #local _rSample = Surcoat_Snow_Samples_Radius;
  #local _hSample = Surcoat_Snow_Samples_Hemisphere;
  #local _minRadius = Surcoat_Snow_MinRadius;
  #local _maxRadius = Surcoat_Snow_MaxRadius;

  //#local _Sky = vnormalize(_DiscNor);
  #local _Right = xVnormalize(vcross(_Sky,x));
  #local _Up = vcross(_Right,_Sky);
  #local _SkyFunction = function{transform{matrix <_Right.x,_Right.y,_Right.z,_Sky.x,_Sky.y,_Sky.z,_Up.x,_Up.y,_Up.z,0,0,0>}}
  
  //File I/O
  #fopen _In _SnowInFile read
  #fopen _Out _SnowOutFile write
  //Save Sky-Vector to file
  #write (_Out, _Sky,"\n")
  #local _Load = 0;  
  //Loop through all particles
  #while (defined(_In))
    //Get data
    #read (_In, _Pos, _Nor)
    #local _Load = _Load + 1;
    #if (Surcoat_Snow_DebugSampleCount>0)
      #if (mod(_Load,Surcoat_Snow_DebugSampleCount)=0 & _Load > 0)
        #debug concat(" -Processed ",str(_Load,0,0)," samples...\n")
      #end
    #end
    #local _Continue = 1;
    //Process Data for particle
    //-Check Max-Height above particle
    //Technique: Shoot upward and memorize distance to first hit.
    #local _tNormal = <0,0,0>;
    #local _upperMaxHeightPos = trace(_Object,_Pos+Surcoat_Sample_Offset*_Nor,_Sky,_tNormal);
    #if (vlength(_tNormal))
      #local _upperMaxHeight = min(vlength(_upperMaxHeightPos-_Pos),Surcoat_Snow_MaxHeight);
    #else
      #local _upperMaxHeight = Surcoat_Snow_MaxHeight;
    #end
    #if (_upperMaxHeight < _minRadius) #local _Continue = 0; #end
    
    //-Check Max-Height below particle
    //Technique: Shoot downward and memorize distance to first hit.
    #if (_Continue)
      #local _tNormal = <0,0,0>;
      #local _lowerMaxHeightPos = trace(_Object,_Pos-Surcoat_Sample_Offset*_Nor,-_Sky,_tNormal);
      #if (vlength(_tNormal))
        #local _lowerMaxHeight = min(vlength(_lowerMaxHeightPos-_Pos),Surcoat_Snow_MaxHeight);
      #else
        #local _lowerMaxHeight = Surcoat_Snow_MaxHeight;
      #end
      #if (_lowerMaxHeight < _minRadius) #local _Continue = 0; #end
    #end

    //-Check Radius
    //Technique: Using the often-used Fibonacci-Distribution, an area around
    // the particle is checked to see how big it can be without floating above
    // the object's surface.
    #if (_Continue)
      #local _RandRot = y*360*rand(_R);
      #local _minMax = <_minRadius,_maxRadius>;
      #local _Sample=0;
      #while (_Sample<_rSample)
        #local _tRadius = _maxRadius * sqrt(_Sample/(_rSample-1));
        #local _FibDist = vrotate(<1,0,0>,y*(360/.618*_Sample)+_RandRot)*_tRadius;
        #local _tNormal = <0,0,0>;
        #local _tStart = _SkyFunction(_FibDist.x,_upperMaxHeight,_FibDist.z)+_Pos;
        #local _radDistance = vlength(trace(_Object,_tStart,-_Sky,_tNormal)-_Pos);
        //Increase min-Radius when position is okay, stop when not okay
        #if (vlength(_tNormal) & _radDistance <= _minMax.v)
          #local _minMax = <_radDistance,_minMax.v>;
        #else
          #local _Sample = _rSample;
        #end
        #local _Sample=_Sample+1;
      #end
      #local _finalRadius = _minMax.u;
      //Adjust Max-Height to percentage of finalRadius vs MaxRadius
      // -this takes care of particles near edges *and* walls, as they would
      //  normally stack due to the walls, but now don't because of the radius
      //  limit.
      #local _upperMaxHeight = _upperMaxHeight * sqrt(_finalRadius/_maxRadius);
      #if (_finalRadius < _minRadius) #local _Continue = 0; #end
    #end

    //-Check Hemispheres for wall-stacking
    //Technique: Sample the hemisphere, add up distances. Average these. If the average
    // distance is close enough, particle will receive upward scaling to simulate stacking.
    #if (_Continue)
      #local _RandRot = y*360*rand(_R);
      #local _distScale = _upperMaxHeight*Surcoat_Snow_WallInfluence;
      #local _minHemi = _distScale;
      #local _addHemi = 0;
      #local _closestHemi = _distScale;
      #local _Sample=0;
      #while (_Sample<_hSample)

        #local _tRadius = sqrt(_Sample/(_hSample-1));
        #local _FibDist = vrotate(<1,0,0>,y*(360/.618*_Sample)+_RandRot)*_tRadius;
        #local _tHeight = sqrt(1-_tRadius*_tRadius);
        #local _tNormal = <0,0,0>;
        
        #ifdef (_NormalF) #undef _NormalF #end
        #local _Right = xVnormalize(vcross(_Nor,x));
        #local _Up = vcross(_Right,_Nor);
        #local _NormalF = function{transform{matrix <_Right.x,_Right.y,_Right.z,_Nor.x,_Nor.y,_Nor.z,_Up.x,_Up.y,_Up.z,0,0,0>}}
        
        #local _tDir = _NormalF(_FibDist.x,_tHeight,_FibDist.z);
        #local _hemiDistance = vlength(trace(_Object,_Pos+_Nor*Surcoat_Sample_Offset,_tDir,_tNormal)-(_Pos+_Nor*Surcoat_Sample_Offset));
        #if (!vlength(_tNormal))
          #local _hemiDistance = _distScale;
        #end        
        //Method 0 : Memorize closest distance
        #if (_hemiDistance < _closestHemi)
          #local _closestHemi = _hemiDistance;
        #end        
        //Method 1 : Add all distances for later averaging
        #local _addHemi = _addHemi+min(_hemiDistance,_distScale);
        #local _Sample=_Sample+1;
      #end
      
      //Simply pick the closest sample and use that as basis for height-modification
      #if (Surcoat_Snow_StackMethod=0)
        #local _upperMaxHeight = _upperMaxHeight * pow((_distScale-_closestHemi)/_distScale,3) * abs(vdot(_Sky,_Nor));
      #end
      //Average all samples and use that as height-modification. To give more weight to closer samples, the average
      // result is multiplied with 3, but clipped at 1.
      #if (Surcoat_Snow_StackMethod=1)
        #local _upperMaxHeight = max(_upperMaxHeight * min((1-_addHemi/_hSample/_distScale)*3,1) * abs(vdot(_Sky,_Nor)),0);
      #end
    #end
    //Save particle if _minRadius isn't underscored
    #if (_Continue)
      #write (_Out, ",",_Pos, ",",_Nor,",",_upperMaxHeight,",",_lowerMaxHeight,",",_finalRadius,"\n")
    #end
  #end  
  #fclose _In
  #fclose _Out
  #if (Surcoat_Debug)
    #debug "Done!\n"
  #end
#end

//Visualizes Processed Particles with Blobbed-Cylinders
#macro Snowfile_Display(_SnowInFile,_Threshold,_Radius,_Strength,_HeightModifier)
  #if (Surcoat_Debug)
    #debug concat("\nDisplaying \"",_SnowInFile,"\"...\n")
  #end
  #fopen _In _SnowInFile read
  #read (_In, _Sky)

  #local _Sph=0;
  #local _Cyl=0;
  blob{threshold _Threshold
    #while (defined(_In))
      #read (_In, _Pos,_Nor,_pHeight,_pBase,_pRadius)
      #if (_pHeight = 0 | _HeightModifier = 0)
        sphere{0,_Radius*_pRadius,_Strength translate _Pos}
        #local _Sph=_Sph+1;
      #else
        #local _modSky = Snowfile_Display_SkyMod(_Sky,_Pos,_Nor);
        cylinder{0,_modSky*max(_pHeight*_HeightModifier,.01),_Radius*_pRadius,_Strength
          translate _Pos
        }
        #local _Cyl=_Cyl+1;
      #end
    #end
  }  
  #if (Surcoat_Debug)
    #debug concat(" -Placed ",str(_Sph,0,0)," spheres and ",str(_Cyl,0,0)," cylinders.\n")
    #debug "Done!\n"
  #end
#end

//Modifies the height of the snow-particles. Negative return-results
// will remove the particle.
#macro Snowfile_Modify(_SnowInFile,_SnowOutFile)
  #if (Surcoat_Debug)
    #debug concat("\nModifying particles from \"",_SnowInFile,"\" and saving results to \"",_SnowOutFile,"\"...\n")  
  #end
  #local _Load=0;
  #local _Save=0;
  //File I/O
  #fopen _In _SnowInFile read
  #fopen _Out _SnowOutFile write
  //Save Sky-Vector to file
  #read (_In, _Sky)
  #write (_Out, _Sky,"\n")
  #while (defined(_In))
    #read (_In, _Pos,_Nor,_Height,_Base,_Radius)
    #local _Load=_Load+1;
    #local _Modifier = Snowdata2Float_HeightMod(_Pos,_Nor,_Radius,_Height);
    #local _Add = Snowdata2Float_HeightAdd(_Pos,_Nor,_Radius,_Height);
    #if (_Modifier >= 0 &
         _Height*_Modifier+_Add >= 0)
      #write (_Out, ",",_Pos, ",",_Nor,",",_Height*_Modifier+_Add,",",_Base,",",_Radius,"\n")
      #local _Save=_Save+1;
    #end
  #end
  #fclose _In
  #fclose _Out
  #if (Surcoat_Debug)
    #debug concat(" -Loaded ",str(_Load,0,0)," particles of which ",str(_Save,0,0)," were modified and saved.\n")
    #debug "Done!\n"
  #end
#end

//Safety-Macro: if x is given direction, vcross(_dir,x)
// would yield 0. This macro returns y.
#macro xVnormalize(_Vector)
  #if (vlength(_Vector))
    #local _Return = vnormalize(_Vector);
  #else
    #local _Return = y;
  #end
  _Return
#end
