// Persistence of Vision Ray Tracer Scene Description File
// File: Surcoat.inc
// Vers: 3.6.1
// Desc: Surcoat-Macros
// Date: 08.12.04 (dd.mm.yy)
// Auth: Tim Nikias Wenclawiak - www.NoLights.de
// Last Update: 24.01.05 (dd.mm.yy)

//Description
// Various macros to sample a surcoat onto an object's surface, consult
// example files and HTML-Help for further information.

//List of Macros
// Surcoat_Orthoview      -> Scene-Macro to create orthographic view of an object
// Surcoat_CreateSamples  -> Samples the Surface of the object

// Surcoat_PrepareSorting -> Prepares the sorting macros
// Surcoat_KeepAngled     -> Keeps samples with normal not exceeding the given limit
// Surcoat_KeepInside     -> Keeps samples inside a given object
// Surcoat_KeepMesh       -> Counts surfaces passing from sample to sky and removes either even or uneven
// Surcoat_RemoveRoofed   -> Removes samples with a roof and thus can't see the sky
// Surcoat_RemoveOnObject -> Removes samples which lie on another object

//Surcoat_SaveSurface     -> Saves unsorted samples to file
//Surcoat_SaveSorted      -> Saves sorted samples to file
//Surcoat_LoadSurface     -> Loads samples from file into unsorted memory
//Surcoat_AppendSamples   -> Combines several sample-files (filenames as strings in 1D-array) into one

//Surcoat_Spheres         -> Places spheres on the samples
//Surcoat_BlobSpheres     -> Creates blob with spheres on the samples

//Surcoat_SortingFileIO_Begin -> Prepares sorted files to be read via macros
//Surcoat_GetSample           -> Gets the next sample in the list
//Surcoat_SetSample           -> Saves the sample to the list
//Surcoat_SortingFileIO_End   -> Cleans up the sorting via macros

//Notes
// Four files will be created during the usage of the macros:
// _ds_prep.tmp
// _ds_seed.tmp
// _ds_sort.tmp
// _ds_thru.tmp
// When sampling an object, the samples are saved to "_ds_seed.tmp". Everytime
// Surcoat_PrepareSorting() is called, the "seeds" get copied to "_ds_prep.tmp".
// Every sorting macro will then read through "prep" and save selected samples
// to "sort", and the macro ends with copying the data from "sort" back to "prep",
// as a preparation for the next sorting macro or saving the sorted samples to
// disk.
// "_ds_thru.tmp" only holds the temporary samples during the shooting of a single
// ray.
// Surcoat_SaveSurface saves content of "_ds_seed.tmp" to disk. LoadSurface copies
// content to "_ds_seed.tmp" and thus can be used to load a set of samples into
// Surcoat's "memory" for multiple sorting using several PrepareSorting() and
// SaveSorted() macro-calls.

#declare _surcoat_inc_tnw = 1;

//Default values
//Topview is not declared as this avoids the pretesting-subroutine. Declaring
// it as a white pigment will start the subroutine, but with the same effect:
// all cells have to be sampled.
//#declare Surcoat_Topview = pigment{rgb 1}
#declare Surcoat_Debug = on;
#declare Surcoat_Seed = 260981;

//The origin of the samples is a rectangle which matches the image-result of
// the Orthoview-macro. The actual unit-size of the object multiplied with
// the desired resolution will define the resolution of the grid layed on top
// of this rectangle. Per cell, the given amount of samples will be shot at
// the object.
#declare Surcoat_CellResolution = 2;
#declare Surcoat_CellSamples = 10;

//The coating is done along a certain direction.
#declare Surcoat_Sample_Direction = -y;
//Samples of which the surface-normal deviates from the given shooting direction
// by more than the given angle (in degrees) will be dropped. 90 ensures that
// every surface up till perpendicularity will get coated.
#declare Surcoat_Sample_MaxDirectionDeviation = 90;
//This will switch off the DirectionDeviation-Check completely and every hit onto
// the object, no matter if top- or bottomside, will be saved.
#declare Surcoat_Sample_OverrideDirection = 0;

//For accuracy reasons, an offset can be defined for the sorting macro, as shooting
// trace()-rays exactly from a sample's position can yield different results (e.g.
// it might hit the exact same surface right away, or pass through and hit the next
// one, depending on transformations and internal representations, to circumvent this,
// the trace()-ray will be offset by the given amount along the normal).
#declare Surcoat_Sample_Offset = .0001;

//Orthoview Macro to create orthographic view of Object
// for pretesting.
#macro Surcoat_Orthoview(_Object)
  #local _Direction = Surcoat_Sample_Direction;
  #local BB_Min = min_extent(_Object);
  #local BB_Max = max_extent(_Object);
  #local _Center = (BB_Min+BB_Max)/2;
  //"World" Axis'
  #local _Sky = xVnormalize(vcross(_Direction,x));
  #local _Right = vnormalize(vcross(_Sky,_Direction));
  #local _Up = vnormalize(_Direction);
  //Matrix Transformation
  #local _Matrix = transform{matrix <_Right.x,_Right.y,_Right.z,_Sky.x,_Sky.y,_Sky.z,_Up.x,_Up.y,_Up.z,0,0,0>}
  #local _BB_Box = object{_Object translate -_Center transform{_Matrix inverse}}
  //Size-Gathering
  #local nBB_Min = min_extent(_BB_Box);
  #local nBB_Max = max_extent(_BB_Box);
  #local fBB_Size = max(nBB_Max.x-nBB_Min.x,nBB_Max.y-nBB_Min.y);
  global_settings{ambient_light 1}
  background{rgb 0}
  camera{
    orthographic
    location -_Up*100
    sky _Sky
    direction _Up
    right _Right*fBB_Size
    up _Sky*fBB_Size
  }
  object{_Object
    material{texture{pigment{rgb 1}finish{ambient 1}}}
    translate -_Center
  }
#end

//Main Macro to coat an object
#macro Surcoat_CreateSamples(_Object)
  #if (Surcoat_Debug)
    #debug "\nDropping samples onto object's surface...\n"
  #end
  #local _Rand = seed(Surcoat_Seed);
  //Calculations for orthographic projection of object/dropping direction
  //Important: These is *exactly* the same routine as used by Surcoat_Orthoview.
  //If only one of both is changed, the Pretest won't match with the final drop.
  #local BB_Min = min_extent(_Object);
  #local BB_Max = max_extent(_Object);
  #local _Center = (BB_Min+BB_Max)/2;
  //"Object" Axis
  #local _Sky = xVnormalize(vcross(Surcoat_Sample_Direction,x));
  #local _Right = vnormalize(vcross(_Sky,Surcoat_Sample_Direction));
  #local _Up = vnormalize(Surcoat_Sample_Direction);
  //Centering etc
  #local _nCenter = _Center.x*_Right+_Center.y*_Sky+_Center.z*_Up;
  #local _Matrix = transform{matrix <_Right.x,_Right.y,_Right.z,_Sky.x,_Sky.y,_Sky.z,_Up.x,_Up.y,_Up.z,0,0,0>}
  #local _BB_Box = object{_Object translate -_Center transform{_Matrix inverse} translate _Center}
  #local nBB_Min = min_extent(_BB_Box);
  #local nBB_Max = max_extent(_BB_Box);
  #local _nCenter = (nBB_Max+nBB_Min)/2;
  #local _xSizeView = nBB_Max.x-nBB_Min.x;
  #local _zSizeView = nBB_Max.y-nBB_Min.y;
  #local fBB_Width = max(_xSizeView,_zSizeView);
  #local fBB_Height = max(_xSizeView,_zSizeView);
  //Macro to convert unit-cube to Orthographic-Space, used in _Drop_Through-Macro
  #macro _Pr(_Vector)
    ((_Vector.x-.5)*_Right*fBB_Width+(_Vector.z-.5)*_Sky*fBB_Height-_Vector.y*_Up*fBB_Height+_Center)
  #end
  //Size of Boundary Box
  #local _BB_Min = min_extent(_Object);
  #local _BB_Max = max_extent(_Object);
  #local _BB_Abs = _BB_Max-_BB_Min;
  #if (Surcoat_Debug)
    #debug concat(" -Boundary-Box Size: <",str(_BB_Abs.x,1,5),",",str(_BB_Abs.y,1,5),",",str(_BB_Abs.z,1,5),">\n")
  #end
  //Default values
  #ifndef (Surcoat_CellResolution)
    #local _PreRes = 5;
  #else
    #local _PreRes = Surcoat_CellResolution;
  #end
  //Pretest does a quick search for empty areas
  #local _Surcoat_Pretest = 1;
  #ifndef (Surcoat_Topview)
    #local _Surcoat_Pretest = 0;
    #local _PA_Res = <ceil(_PreRes*fBB_Width),ceil(_PreRes*fBB_Height)>;
    #if (Surcoat_Debug)
      #debug concat(" -No Topview declared, pretest not done, prepared ",str(_PA_Res.u*_PA_Res.v,0,0)," cells...\n")
    #end
    #local _PreArray = array[_PA_Res.u][_PA_Res.v];
    #local _A=0;
    #while (_A<_PA_Res.u)
      #local _B=0;
      #while (_B<_PA_Res.v)
        #local _PreArray[_A][_B] = 11;
        #local _B=_B+1;
      #end
      #local _A=_A+1;
    #end
    #local _EmptyCells = 0;
  #end
  //Do a subdivision of the area to check for empty
  //spaces which need no further testing.
  #if (_Surcoat_Pretest)
    //Function for the Topview-Image
    #local _TVI_t = function{pigment{Surcoat_Topview rotate x*180}};
    #local _TVI = function{_TVI_t(x,y,z).x};
    //Array to save which cells are active
    #local _PA_Res = <ceil(_PreRes*fBB_Width),ceil(_PreRes*fBB_Height)>;
    #local _PreArray = array[_PA_Res.u][_PA_Res.v];
    #if (Surcoat_Debug)
      #debug concat(" -Prepared ",str(_PA_Res.u*_PA_Res.v,0,0)," cells...\n")
    #end
    //Generate on/off values for each cell
    /*Technique description:
      The function from the orthoview image is used to create a heightfield triple
      the resolution of the cells. Then, along the edges of the cells, I shoot trace
      to see if I hit the heightfield. If I do, the object somehow reaches into/out
      of the cell and thus needs to be active. If the trace() yields no result on all
      edges, I do a check if the cell is inside the heightfield, if so, the cell is
      active, otherwise it isn't.
      At the end, every checked cell receives a value (>0 = active). The value is
      deriven from the edge on which the check was successful.
      
      For visualization purposes: the array is x+z, first row is the row with the
      largest z-value in POV-space, last row the one with the lowest z-value.
    */
    #local _maxTraceDist = <1/_PA_Res.u,1/_PA_Res.v>;
    #local _Displace = 1-_maxTraceDist.u*.01;
    #local _TV_HF = 
      height_field{
        function _PA_Res.u*3,_PA_Res.v*3 {_TVI((x-.5)*_Displace+.5,(y-.5)*_Displace+.5,(z-.5)*_Displace+.5)}
      }
    #local _EmptyCells = 0;
    #local _A=0;
    #while (_A<_PA_Res.u)
      #local _Xc = (_A/_PA_Res.u);
      #local _B=0;
      #while (_B<_PA_Res.v)
        #local _Zc = (_B/_PA_Res.v);
        //Initialize with 1, if still 1 at end, set to 0
        #local _PreArray[_A][_B] = 1;
        //Top edge
        #local _N = <0,0,0>;
        #local _Start = <_Xc,.05,1-_Zc>;
        #local _Dir = x;
        #if (vlength(trace(_TV_HF,_Start,_Dir,_N)-_Start) < _maxTraceDist.u &
             vlength(_N))
          #local _PreArray[_A][_B] = _PreArray[_A][_B]*2;
        #end
        //Right edge
        #local _N = <0,0,0>;
        #local _Start = <_Xc+_maxTraceDist.u,.05,1-_Zc>;
        #local _Dir = -z;
        #if (vlength(trace(_TV_HF,_Start,_Dir,_N)-_Start) < _maxTraceDist.v &
             vlength(_N))
          #local _PreArray[_A][_B] = _PreArray[_A][_B]*3;
        #end
        //Bottom edge
        #local _N = <0,0,0>;
        #local _Start = <_Xc,.05,1-_Zc-_maxTraceDist.v>;
        #local _Dir = x;
        #if (vlength(trace(_TV_HF,_Start,_Dir,_N)-_Start) < _maxTraceDist.u &
             vlength(_N))
          #local _PreArray[_A][_B] = _PreArray[_A][_B]*5;
        #end
        //Left edge
        #local _N = <0,0,0>;
        #local _Start = <_Xc,.05,1-_Zc>;
        #local _Dir = -z;
        #if (vlength(trace(_TV_HF,_Start,_Dir,_N)-_Start) < _maxTraceDist.v &
             vlength(_N))
          #local _PreArray[_A][_B] = _PreArray[_A][_B]*7;
        #end
        //If no edge was hit:
        #if (_PreArray[_A][_B] = 1)
          #if (inside(_TV_HF,<_Xc,.5,1-_Zc>+(x*_maxTraceDist.u-z*_maxTraceDist.v)*.5))
            #local _PreArray[_A][_B] = 11;
          #end
        #end
        //If no edge was hit and the center isn't inside the heightfield,
        //the cell is considered empty and set to 0
        #if (_PreArray[_A][_B] = 1)
          #local _PreArray[_A][_B]=0;
          #local _EmptyCells = _EmptyCells+1;
        #end
        #local _B=_B+1;
      #end
      #local _A=_A+1;
    #end
  #end
  #if (Surcoat_Debug)
    #debug concat(" -",str(_EmptyCells,0,0)," cells are empty.\n")
  #end
  /*Technique description
    From here on, we have an array which tells us active and inactive cells.
    The activation-number can also be used (via prime factorization) to find
    out, which edges contributed to the active-state. Generally, an active
    cell will just be filled with a set amount of particles. I had in mind of
    doing an additional drop along the edges if the first attempt fails, but
    I'd rather have even distribution along with the problems of discretization
    (too few samples may return incorrect results) than edges showing up now
    and then. Additionally, this macro thus turn into a possibility to just 
    use the raw-data instead of *requiring* a clean-up pass.
    
    To drop a particle, I call the _Drop_Through macro, it returns a value
    of how many positions *on* the surfaces it encountered. These were saved
    to "_ds_thru.tmp" during the execution of the macro.
    All these positions then get added to "_ds_seed.tmp". The amount of total
    positions gets tracked and will be saved to the specified filename at the
    end of this macro.
  */
  #ifndef (Surcoat_CellSamples)
    #local _PperCell = 25;
  #else
    #local _PperCell = Surcoat_CellSamples;
  #end
  #local _PA_Res = <dimension_size(_PreArray,1),dimension_size(_PreArray,2)>;
  #local _maxTraceDist = <1/_PA_Res.u,1/_PA_Res.v>;
  #if (Surcoat_Debug)
    #debug concat(" -Commencing sampling: Attempting to take ",str((_PA_Res.u*_PA_Res.v-_EmptyCells)*_PperCell,0,0)," samples...\n")
  #end
  //Highest Radius for cells
  #local _MaxRad = sqrt(pow(1/_PreRes*.5,2)*2)*1.05;
  #local _TotalHits = 0;
  #local _ParticlesDropped = 0;
  #local _A=0;
  #while (_A<_PA_Res.u)
    #local _Xc = _A/_PA_Res.u;
    #local _B=0;
    #while (_B<_PA_Res.v)
      #local _Zc = _B/_PA_Res.v;
      #local _RandRot = rand(_Rand)*360;
      #if (_PreArray[_A][_B]!=0)
        #local _ParticlesDropped = _ParticlesDropped + _PperCell;
        //Center of cell
        #local _StartPosCenter = <_Xc+_maxTraceDist.u*.5,0,1-_Zc-_maxTraceDist.v*.5>+y*(_BB_Max-_BB_Min+1);
        #local _C=0;
        #while (_C<_PperCell)
          //Fibonacci-Distribution
          #local _FibDist = vrotate(<_MaxRad,0,0>,y*(360/.618*_C+_RandRot)) * sqrt(_C/(_PperCell-1));
          //Scale FibDist adjust to the later projection onto the orthoview-plane
          #local _FibDist = _FibDist/fBB_Height;
          //Only do stuff with positions *inside* the cell
          #if ( abs(_FibDist.x)<=1/_PreRes &
                abs(_FibDist.z)<=1/_PreRes)
            #local _StartPos = _StartPosCenter+_FibDist;
            #local _Hits = _Drop_Through();
          #else
            #local _Hits = 0;
          #end
          //And now save hits to "_ds_seed.tmp"
          #if (_Hits!=0)
            _AccumulateSeeds(_TotalHits)
          #end
          #local _C=_C+1;
        #end // of _C-Loop
      #end // of PreArray-Active check
      #local _B=_B+1;
    #end // of _B-Loop
    #local _A=_A+1;
  #end // of _A-Loop
  #if (Surcoat_Debug)
    #debug concat(" -Traced ",str(_ParticlesDropped,0,0)," individual paths and saved ",str(_TotalHits,0,0)," samples.\n")
  #end
  //Clear temporary files of data
  #fopen _DThru "_ds_thru.tmp" write
  #fclose _DThru
  //Undeclare Orthoview-Pigment if declared
  #ifdef (Surcoat_Topview) #undef Surcoat_Topview #end
  #if (Surcoat_Debug)
    #debug concat("Done!\n")
  #end
#end


//Aid-Macros to workaround specific problems or for ease of use
//=============================================================

//Safety-Macro: if x is given direction, vcross(_dir,x)
// would yield 0. This macro returns y.
#macro xVnormalize(_Vector)
  #if (vlength(_Vector)) #local _Return = vnormalize(_Vector); #else #local _Return = y; #end
  _Return
#end

//Macro which shoots through the object, saves hits along
// with normal to "_ds_thru.tmp", and returns number for
// amount of hits.
#macro _Drop_Through()
  //position and object defined in local layer
  #local _Return = 0;
  #fopen _DThru "_ds_thru.tmp" write
  //Loop which shoots through the object
  #local _N = <0,0,0>;
  #local _St = _Pr(_StartPos);
  #local _End = trace(_Object,_St,Surcoat_Sample_Direction,_N);
  #while (vlength(_N))
    //Only upwards-pointing normals are accepted (or if the test is overridden, every surface)
    #if (vdot(_N,-_Up) > cos(radians(Surcoat_Sample_MaxDirectionDeviation)) | Surcoat_Sample_OverrideDirection)
      #if (_Return)
        #write (_DThru, ",",_End,",",_N,"\n")
      #else
        #write (_DThru, _End,",",_N,"\n")
      #end
      #local _Return = _Return+1;
    #end
    #local _End = trace(_Object,_End,Surcoat_Sample_Direction,_N);
  #end
  //Close temporary file  
  #fclose _DThru
  //Return number of hits
  _Return
#end

//Collects Hits from _Drop_Through into "_ds_seed.tmp"
#macro _AccumulateSeeds(_TotalHits)
  #if (_TotalHits)
    #fopen _Surcoat2File "_ds_seed.tmp" append
  #else
    #fopen _Surcoat2File "_ds_seed.tmp" write
  #end
  #fopen _Surcoat3File "_ds_thru.tmp" read
  #local _A=0;
  #while (_A<_Hits)
    #read (_Surcoat3File, _Pos,_Nor)
    #if (_TotalHits)
      #write (_Surcoat2File, ",", _Pos, ",", _Nor,"\n")
    #else
      #write (_Surcoat2File, _Pos, ",", _Nor,"\n")
    #end  
    #declare _TotalHits = _TotalHits+1;
    #local _A=_A+1;  
  #end
  #fclose _Surcoat2File
  #fclose _Surcoat3File
#end

//Collects Particles from _Inni and drops them into _Outi
#macro _Surcoat_File2File(_Inni,_Outi)
  #fopen _In _Inni read
  #fopen _Out _Outi write
  #local _Load = 0;
  #local _Save = 0;
  #while (defined(_In))
    #read (_In, _Pos, _Nor)
    #local _Load=_Load+1;
    #if (_Save=0)
      #write (_Out, _Pos,"," _Nor,"\n")
    #else
      #write (_Out, ",",_Pos,"," _Nor,"\n")
    #end
    #local _Save=_Save+1;
  #end
  #fclose _In
  #fclose _Out
#end


//Sorting Macros
//==============

//Sorting preparation
#macro Surcoat_PrepareSorting()
  #if (Surcoat_Debug)
    #debug "\nPreparing Sample-Selecting...\n"
   #end
  _Surcoat_File2File("_ds_seed.tmp","_ds_prep.tmp")
  #if (Surcoat_Debug)
    #debug "Done!\n"
   #end
#end

//Parses through the prepared seeds and removes those
// which have a larger angle to the dropping direction
// than specified.
#macro Surcoat_KeepAngled(_Dir,_Angle)
  #if (Surcoat_Debug)
    #debug "\nChecking samples on angled surfaces...\n"
  #end
  #fopen _In "_ds_prep.tmp" read
  #fopen _Out "_ds_sort.tmp" write
  #local _Load = 0;
  #local _Save = 0;
  #while (defined(_In))
    #read (_In, _Pos, _Nor)
    #local _Load=_Load+1;
    #local _AngledResult = degrees(acos(vdot(_Nor,vnormalize(_Dir))));
    #if (_AngledResult<=_Angle)
      #if (_Save=0)
        #write (_Out, _Pos,"," _Nor,"\n")
      #else
        #write (_Out, ",",_Pos,"," _Nor,"\n")
      #end
      #local _Save=_Save+1;
    #end
  #end
  #if (Surcoat_Debug)
    #debug concat(" -Loaded ",str(_Load,0,0)," samples and kept ",str(_Save,0,0),".\n")
  #end
  #fclose _In
  #fclose _Out
  //Copy sorted to prep for next Macro
  _Surcoat_File2File("_ds_sort.tmp","_ds_prep.tmp")
#end

//Keeps particles inside the object, makes use of
// inside() float-function
#macro Surcoat_KeepInside(_Object)
  #if (Surcoat_Debug)
    #debug "\nChecking samples within object...\n"
  #end
  #fopen _In "_ds_prep.tmp" read
  #fopen _Out "_ds_sort.tmp" write
  #local _Load = 0;
  #local _Save = 0;
  #while (defined(_In))
    #read (_In, _Pos, _Nor)
    #local _Load=_Load+1;
    #if (inside(_Object,_Pos))
      #if (_Save=0)
        #write (_Out, _Pos,"," _Nor,"\n")
      #else
        #write (_Out, ",",_Pos,"," _Nor,"\n")
      #end
      #local _Save=_Save+1;
    #end
    
  #end
  #if (Surcoat_Debug)
    #debug concat(" -Loaded ",str(_Load,0,0)," samples and kept ",str(_Save,0,0),".\n")
  #end
  #fclose _In
  #fclose _Out
  //Copy sorted to prep for next Macro
  _Surcoat_File2File("_ds_sort.tmp","_ds_prep.tmp")
#end

//Keeps particles which lie on in- or outside of mesh, based
// upon how many surfaces a ray passes through when reaching
// the particle's position from the position outward.
// 0 = Outside, 1= Inside
#macro Surcoat_KeepMesh(_Mesh,_ModFactor)
  #if (Surcoat_Debug)
    #debug "\nChecking samples for inside/outside of mesh...\n"
  #end
  #fopen _In "_ds_prep.tmp" read
  #fopen _Out "_ds_sort.tmp" write
  #local _Load = 0;
  #local _Save = 0;
  #while (defined(_In))
    #read (_In, _Pos, _Nor)
    #local _Load=_Load+1;
    #local _CountedSurfaces = 0;
    #local _Temp = _Pos+_Nor*Surcoat_Sample_Offset;
    #local _TempNormal = <0,0,0>;
    #local _Temp = trace(_Mesh,_Temp,-_Nor,_TempNormal);
    #while (vlength(_TempNormal))
      #local _CountedSurfaces = _CountedSurfaces + 1;
      #local _TempNormal = <0,0,0>;
      #local _Temp = trace(_Mesh,_Temp,-_Nor,_TempNormal);
    #end
    #if (mod(_CountedSurfaces,2)=_ModFactor)
      #if (_Save=0)
        #write (_Out, _Pos,"," _Nor,"\n")
      #else
        #write (_Out, ",",_Pos,"," _Nor,"\n")
      #end
      #local _Save=_Save+1;
    #end
  #end
  #if (Surcoat_Debug)
    #debug concat(" -Loaded ",str(_Load,0,0)," sampled and kept ",str(_Save,0,0),".\n")
  #end
  #fclose _In
  #fclose _Out
  //Copy sorted to prep for next Macro
  _Surcoat_File2File("_ds_sort.tmp","_ds_prep.tmp")
#end

//Removes particles which are "roofed": when a certain disc above
// the particle isn't clear enough, the particle is removed. Quite
// parsing-intensive script as it requires samples*particles trace-
// calls.
#macro Surcoat_RemoveRoofed(_Object,_DiscDist,_DiscNor,_DiscRad,_DiscPercent,_Samples)
  #if (Surcoat_Debug)
    #debug "\nRemoving samples with \"roof\"...\n"
  #end
  #fopen _In "_ds_prep.tmp" read
  #fopen _Out "_ds_sort.tmp" write
  #local _Load = 0;
  #local _Save = 0;
  #local _R = seed(Surcoat_Seed);
  #if (vlength(_DiscNor))
    #local _Sky = vnormalize(_DiscNor);
    #local _Right = xVnormalize(vcross(_Sky,x));
    #local _Up = vcross(_Right,_Sky);
    #local _NormalF = function{transform{matrix <_Right.x,_Right.y,_Right.z,_Sky.x,_Sky.y,_Sky.z,_Up.x,_Up.y,_Up.z,0,0,0>}}
    #local _PreMatrix = 1;
  #else
    #local _PreMatrix = 0;
  #end
  #while (defined(_In))
    #read (_In, _Pos, _Nor)
    #local _Load=_Load+1;
    //Generate Matrix for Disc
    #if (!_PreMatrix)
      #ifdef (_NormalF) #undef _NormalF #end
      #local _Sky = vnormalize(_Nor);
      #local _Right = xVnormalize(vcross(_Sky,x));
      #local _Up = vcross(_Right,_Sky);
      #local _NormalF = function{transform{matrix <_Right.x,_Right.y,_Right.z,_Sky.x,_Sky.y,_Sky.z,_Up.x,_Up.y,_Up.z,0,0,0>}}
    #end
    //Generate Disc-Positions
    #local _TooClose = 0;
    #local _Sample=0;
    //disc{_DiscNor*_DiscDist,_DiscNor,_DiscRad pigment{rgb y} transform{matrix <_Right.x,_Right.y,_Right.z,_Sky.x,_Sky.y,_Sky.z,_Up.x,_Up.y,_Up.z,0,0,0>} translate _Pos}
    #local _RandRot = y*360*rand(_R);
    #while (_Sample<_Samples)
      //Fibonacci-Distribution
      #local _tRadius = _DiscRad * sqrt(_Sample/(_Samples-1));
      #local _FibDist = vrotate(<1,0,0>,y*(360/.618*_Sample)+_RandRot)*_tRadius;
      #local _tNormal = <0,0,0>;
      #local _tEnd = _NormalF(_FibDist.x,_DiscDist,_FibDist.z);
      #local _RoofDistance = vlength(trace(_Object,_Pos+_Nor*Surcoat_Sample_Offset,_tEnd,_tNormal)-(_Pos+_Nor*Surcoat_Sample_Offset));
      #if (vlength(_tNormal) & _RoofDistance <= sqrt(_tRadius*_tRadius+_DiscDist*_DiscDist))
        #local _TooClose = _TooClose + 1;
      #end
      #local _Sample=_Sample+1;
    #end
    #local _PercOkay = _DiscPercent - (_TooClose/(_Samples-1));
    #if (_PercOkay>=0)
      #if (_Save=0)
        #write (_Out, _Pos,"," _Nor,"\n")
      #else
        #write (_Out, ",",_Pos,"," _Nor,"\n")
      #end
      #local _Save=_Save+1;
    #end
  #end
  #if (Surcoat_Debug)
    #debug concat(" -Loaded ",str(_Load,0,0)," samples and kept ",str(_Save,0,0),".\n")
  #end
  #fclose _In
  #fclose _Out
  //Copy sorted to prep for next Macro
  _Surcoat_File2File("_ds_sort.tmp","_ds_prep.tmp")
  
#end

//Removes samples which touch a given object
#macro Surcoat_RemoveOnObject(_Object,_Radius)
  #if (Surcoat_Debug)
    #debug "\nChecking if samples touch an object...\n"
  #end
  #fopen _In "_ds_prep.tmp" read
  #fopen _Out "_ds_sort.tmp" write
  #local _Load = 0;
  #local _Save = 0;
  #while (defined(_In))
    #read (_In, _Pos, _Nor)
    #local _Load = _Load + 1;
    #local _Temp = _Pos+_Nor*Surcoat_Sample_Offset;
    #local _TempNormal = <0,0,0>;
    #local _Temp = trace(_Object,_Temp,-_Nor,_TempNormal);
    
    #if (!vlength(_TempNormal) | vlength(_Temp-_Pos) > _Radius)
      #if (_Save=0)
        #write (_Out, _Pos,"," _Nor,"\n")
      #else
        #write (_Out, ",",_Pos,"," _Nor,"\n")
      #end
      #local _Save=_Save+1;
    #end
    
  #end
  #if (Surcoat_Debug)
    #debug concat(" -Loaded ",str(_Load,0,0)," sampled and kept ",str(_Save,0,0),".\n")
  #end
  #fclose _In
  #fclose _Out
  //Copy sorted to prep for next Macro
  _Surcoat_File2File("_ds_sort.tmp","_ds_prep.tmp")
#end

//File I/O Macros
//===============

//Saves Current Surfaces Samples to given file
#macro Surcoat_SaveSurface(_SnowOutFilename)
  #if (Surcoat_Debug)
    #debug concat("\nSaving \"memory\" samples to \"",_SnowOutFilename,"\"...\n")
  #end
  #fopen _In "_ds_seed.tmp" read
  #fopen _Out _SnowOutFilename write
  #local _Load = 0;
  #local _Save = 0;
  #while (defined(_In))
    #read (_In, _Pos, _Nor)
    #local _Load=_Load+1;
    #if (_Save=0)
      #write (_Out, _Pos,"," _Nor,"\n")
      #local _Save=_Save+1;
    #else
      #write (_Out, ",",_Pos,"," _Nor,"\n")
    #end
  #end
  #if (Surcoat_Debug)
    #debug concat(" -Saved ",str(_Load,0,0)," samples.\n")
  #end
  #fclose _In
  #fclose _Out
#end

//Saves Current Sorted Samples to given file
#macro Surcoat_SaveSorted(_SnowOutFilename)
  #if (Surcoat_Debug)
    #debug concat("\nSaving sorted samples to \"",_SnowOutFilename,"\"...\n")
  #end
  #fopen _In "_ds_prep.tmp" read
  #fopen _Out _SnowOutFilename write
  #local _Load = 0;
  #local _Save = 0;
  #while (defined(_In))
    #read (_In, _Pos, _Nor)
    #local _Load=_Load+1;
    #if (_Save=0)
      #write (_Out, _Pos,"," _Nor,"\n")
      #local _Save=_Save+1;
    #else
      #write (_Out, ",",_Pos,"," _Nor,"\n")
    #end
  #end
  #if (Surcoat_Debug)
    #debug concat(" -Saved ",str(_Load,0,0)," samples.\n")
  #end
  #fclose _In
  #fclose _Out
#end

//Saves Current Surfaces Samples to given file
#macro Surcoat_LoadSurface(_SnowInFilename)
  #if (Surcoat_Debug)
    #debug concat("\nLoading samples from \"",_SnowInFilename,"\" into Surcoat's \"memory\"...\n")
  #end
  #fopen _In _SnowInFilename read
  #fopen _Out "_ds_seed.tmp" write
  #local _Load = 0;
  #local _Save = 0;
  #while (defined(_In))
    #read (_In, _Pos, _Nor)
    #local _Load=_Load+1;
    #if (_Save=0)
      #write (_Out, _Pos,"," _Nor,"\n")
      #local _Save=_Save+1;
    #else
      #write (_Out, ",",_Pos,"," _Nor,"\n")
    #end
  #end
  #if (Surcoat_Debug)
    #debug concat(" -Loaded ",str(_Load,0,0)," samples.\n")
  #end
  #fclose _In
  #fclose _Out
#end

//Appends array to file
#macro Surcoat_AppendSamples(_SnowInArray,_SnowOutFilename)
  #if (Surcoat_Debug)
    #debug concat("\nAppending samples to \"",_SnowOutFilename,"\"...\n")
  #end
  #fopen _Out _SnowOutFilename write
  #local _Size = dimension_size(_SnowInArray,1);
  #local _Save = 0;
  #local _A=0;
  #while (_A<_Size)
  
  #fopen _In _SnowInArray[_A] read
  #while (defined(_In))
    #read (_In, _Pos, _Nor)
    #if (_Save=0)
      #write (_Out, _Pos,"," _Nor,"\n")
      #local _Save=_Save+1;
    #else
      #write (_Out, ",",_Pos,"," _Nor,"\n")
     #local _Save=_Save+1;
   #end
  #end
  #fclose _In
  #local _A=_A+1;
  #end
  #if (Surcoat_Debug)
    #debug concat(" -Saved ",str(_Save,0,0)," samples.\n")
  #end
  #fclose _Out
#end


//Visualization Macros
//====================

//Parsing through each particle and saving properties
#macro Surcoat_Spheres(_CoatInFile,_Radius)
  #if (Surcoat_Debug)
    #debug concat("\nDisplaying \"",_CoatInFile,"\" with spheres...\n")
  #end
  #local _Load=0;
  #fopen _In _CoatInFile read
  //Loop through all particles
  union{
    #while (defined(_In))
      //Get data
      #read (_In, _Pos, _Nor)
      #local _Load = _Load + 1;
      sphere{_Pos,_Radius}
    #end
  }
  #fclose _In
  #if (Surcoat_Debug)
    #debug concat(" -Placed ",str(_Load,0,0)," spheres.\n")
    #debug "Done!\n"
  #end
#end

//Parsing through each particle and saving properties
#macro Surcoat_BlobSpheres(_CoatInFile,_Threshold,_Radius,_Strength)
  #if (Surcoat_Debug)
    #debug concat("\nDisplaying \"",_CoatInFile,"\" with blobbed spheres...\n")
  #end
  #local _Load=0;
  #fopen _In _CoatInFile read
  //Loop through all particles
  blob{threshold _Threshold
    #while (defined(_In))
      //Get data
      #read (_In, _Pos, _Nor)
      #local _Load = _Load + 1;
      sphere{_Pos,_Radius,_Strength}
    #end
  }
  #fclose _In
  #if (Surcoat_Debug)
    #debug concat(" -Placed ",str(_Load,0,0)," spheres.\n")
    #debug "Done!\n"
  #end
#end

//Building Blocks for own sorting macros
//======================================

#macro Surcoat_SortingFileIO_Begin()
  #fopen _tnw_SurcoatSortIn "_ds_prep.tmp" read
  #fopen _tnw_SurcoatSortOut "_ds_sort.tmp" write
  #declare _tnw_SurcoatSortBB_Count = 0;
  #declare Surcoat_SampleLoop = 1;
  #if (Surcoat_Debug)
    #debug "-Prepared Surcoat-File-I/O...\n"
  #end
#end

#macro Surcoat_GetSample(_Position,_Normal)
  #if (defined(_tnw_SurcoatSortIn) & Surcoat_SampleLoop)
    #read (_tnw_SurcoatSortIn, _tnw_somePos, _tnw_someNor)
    #declare _Position = _tnw_somePos;
    #declare _Normal = _tnw_someNor;
    #undef _tnw_somePos
    #undef _tnw_someNor
    #if (!defined(_tnw_SurcoatSortIn))
      #declare Surcoat_SampleLoop = 0;
    #end
  #else
    #declare _Position = <0,0,0>;
    #declare _Normal = <0,0,0>;
    #declare Surcoat_SampleLoop = 0;
  #end
#end

#macro Surcoat_SetSample(_Position,_Normal)
  #if (defined(_tnw_SurcoatSortOut))
    #if (_tnw_SurcoatSortBB_Count=0)
      #write (_tnw_SurcoatSortOut, _Position,"," _Normal,"\n")
    #else
      #write (_tnw_SurcoatSortOut, ",",_Position,"," _Normal,"\n")
    #end
    #declare _tnw_SurcoatSortBB_Count = _tnw_SurcoatSortBB_Count+1;
  #end
#end

#macro Surcoat_SortingFileIO_End()
  #ifdef (_tnw_SurcoatSortIn) #fclose _tnw_SurcoatSortIn #end
  #ifdef (_tnw_SurcoatSortOut) #fclose _tnw_SurcoatSortOut #end
  _Surcoat_File2File("_ds_sort.tmp","_ds_prep.tmp")
  #if (Surcoat_Debug)
    #debug concat("-Finished Surcoat-File-I/O, saved ",str(_tnw_SurcoatSortBB_Count,0,0)," samples!\n")
  #end
  #ifdef (_tnw_SurcoatSortBB_Count) #undef _tnw_SurcoatSortBB_Count #end
#end
